# collections
from .collections import get_sales as get_collection_sales

# policies
from .policies import get_listings as get_policy_listings

# aliases
get_collection_listings = get_policy_listings
